/*
 * Copyright (c) 2004-2006  
 *              Martin Reinecke (1), Klaus Dolag (1)
 *               (1) Max-Plank-Institute for Astrophysics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include<iostream>
#include<cmath>
#include<fstream>
#include<algorithm>

#include "arr.h"
#include "cxxutils.h"
#include "paramfile.h"
#include "kernel/bstream.h"
#include "kernel/colour.h"
#include "config/config.h"
#include "utils/colourmap.h"

using namespace std;
using namespace RAYPP;

#define NO_SORTBYHSML
#define NO_SORTBYVAL
#define NO_REVERSESORT
#define NO_STOP_AFTER_N 5
#define NO_FULLABSROB
#define NO_MINHSMLPIXEL
#define NO_TIPSY
#define NO_PROJECTION_OFF
#define NO_PATCH_MACH
#define GEOMETRY_FILE

const float8 Pi=3.14159265358979323846;

struct PARTICLE
  {
  float4 x,y,z,r,ro,I,T;
  int type;
  };

struct zcmp
  {
  int operator()(const PARTICLE &p1, const PARTICLE &p2)
    {
#ifdef SORTBYVAL
#ifdef REVERSESORT
    return p1.T<p2.T;
#else
    return p1.T>p2.T;
#endif
#else
#ifdef SORTBYHSML
    return p1.r>p2.r;
#else
    return p1.z>p2.z;
#endif
#endif
    }
  };

class exptable
  {
  private:
    int nexp;
    float8 maxexptab;
    float8 expfac;
    float8 *exptab;

  public:
    exptable (int nexp_, float8 maxexp_)
      : nexp(nexp_), maxexptab(maxexp_), expfac(nexp/maxexptab),
        exptab(new float8[nexp])
      {
      for (int m=0; m<nexp; ++m)
        exptab[m]=exp(m/expfac);
      }

    float8 operator() (float8 arg) const
      {
      arg *= expfac;
      int iarg=int(arg);
      if (iarg<0) return 1.;
      if (iarg>=nexp-1) return 0.;
      float8 frac=arg-iarg;
      return (1.-frac)*exptab[iarg]+frac*exptab[iarg+1];
      }
  };

int gadget_find_block(bifstream &file,const string &label)
  {
  int4 blocksize=0, blksize;
  char blocklabel[5]={"    "};

  file.rewind();

  while(!file.eof() && blocksize == 0)
    {
    file >> blksize;
    if (blksize != 8)
      {
	cout << "wrong structure in GADGET file: " << blksize << endl;
      exit(1);
      }
    else
      {
      file.get(blocklabel,4);
      blocklabel[4] = 0;
      //cout << "blocklabel: " << blocklabel << endl;
      file >> blocksize;
      //cout << "blocksize: " << blocksize << endl;
      file.skip(4);
      if (label!=blocklabel)
	{
        file.skip(blocksize);
        blocksize=0;
        }
      }
    }
  return(blocksize-8);
  }

int gadget_read_header(bifstream &file, int *npart,double *massarr,
  double &time,double &redshift, int *npartall)
  {
  int blocksize = gadget_find_block (file,"HEAD");
  planck_assert (blocksize>0, "Header block not found");
  file.skip(4);
  file.get(npart,6);
  file.get(massarr,6);
  file >> time >> redshift;
  file.skip(8);
  file.get(npartall,6);

  //  cout << " HEADER " << npart[0]<< " , " << npartall[0] << " Patching !!" << endl;  
  //  npartall[0]=34012224;
  //  npartall[1]=34012224;
  return blocksize;
  }

int main (int argc, const char ** argv)
  {
const float8 Pi=3.14159265358979323846;
  announce ("splotch");
  planck_assert (argc==2,"usage: splotch <parameter file>");
  paramfile params (argv[1]);

  cout << "initializing" << endl;
  int res = params.find<int>("resolution",200);
  int ycut0 = params.find<int>("ycut0",0);
  int ycut1 = params.find<int>("ycut1",res);
  string infilename = params.find<string>("infile");
  string outfile = params.find<string>("outfile");
  string filename;
  int numfiles = params.find<int>("numfiles",1);
  bool doswap = params.find<bool>("swap_endian",true);

  int lable_int_nr = params.find<int>("lable_intensity_nr",2);
  int lable_col_nr = params.find<int>("lable_color_nr",1);
  string lable_int,lable_col;
  switch(lable_int_nr)
    {
    case 1:lable_int="TEMP"; break;
    case 2:lable_int="RHO "; break;
    case 3:lable_int="U   "; break;
    case 4:lable_int="MACH"; break;
#ifdef PATCH_MACH
    case 5:lable_int="DTEG"; break;
#else
    case 5:lable_int="DISS"; break;
#endif
    }
  switch(lable_col_nr)
    {
    case 1:lable_col="TEMP"; break;
    case 2:lable_col="RHO "; break;
    case 3:lable_col="U   "; break;
    case 4:lable_col="MACH"; break;
#ifdef PATCH_MACH
    case 5:lable_int="DTEG"; break;
#else
    case 5:lable_col="DISS"; break;
#endif
    }
  int colors = params.find<int>("colors",1);
  bool log_int = params.find<bool>("log_intensity",true);
  bool log_col = params.find<bool>("log_color",true);
  float minval_int = params.find<float>("min_int",1024);
  float maxval_int = params.find<float>("max_int",1024);
  float minval_col = params.find<float>("min_col",1024);
  float maxval_col = params.find<float>("max_col",1024);
  float hsml_fac = params.find<float>("hsml_fac",1.0);

  bool stars = params.find<bool>("stars",false);
  float star_hsml = params.find<float>("star_hsml",5.0);
  int stars_int_nr = params.find<int>("stars_intensity_nr",2);
  int stars_col_nr = params.find<int>("stars_color_nr",1);
  string stars_int,stars_col;
  switch(stars_int_nr)
    {
    case 1:stars_int="AGE "; break;
    }
  switch(stars_col_nr)
    {
    case 1:stars_col="AGE "; break;
    }
  bool log_stars_int = params.find<bool>("log_stars_intensity",true);
  bool log_stars_col = params.find<bool>("log_stars_color",true);
  float minval_stars_int = params.find<float>("min_stars_int",1024);
  float maxval_stars_int = params.find<float>("max_stars_int",1024);
  float minval_stars_col = params.find<float>("min_stars_col",1024);
  float maxval_stars_col = params.find<float>("max_stars_col",1024);
  
  float8 brightness = params.find<double>("brightness",1.);

  exptable xexp(10000,-20.);

  int4 npart[6],npartall[6];
  float8 massarr[6],time,redshift;

  if(numfiles>1) filename=infilename+"."+dataToString(0);
  else           filename=infilename;
  bifstream infile(filename.c_str(),doswap);
  if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
  gadget_read_header(infile,npart,massarr,time,redshift,npartall);
  if(!stars) npartall[4]=0;
  cout <<  time << " " << redshift << endl;
  infile.close();

  int np=npartall[0];
  int nstar=npartall[4];
  cout << "Allocating space for " << np << " gas particles and " << nstar 
       << " star particles ..." << endl;  
  PARTICLE *p=new PARTICLE[np+nstar];
#ifdef GEOMETRY_FILE
  PARTICLE *p_orig=new PARTICLE[np+nstar];
#endif
  float4 mintmp=1e30, maxtmp=-1e30,minint=1e30, maxint=-1e30;
  float4 mintmp_stars=1e30, maxtmp_stars=-1e30,minint_stars=1e30, maxint_stars=-1e30;
  int nskip,npstart=0,nstarstart=np;

  for(int f=0;f<numfiles;f++)
    {
     if(numfiles>1) filename=infilename+"."+dataToString(f);
     else           filename=infilename;
     bifstream infile(filename.c_str(),doswap);
     if (!infile) {cout << "could not open input file! <" << filename << ">" << endl; exit(1);}
     gadget_read_header(infile,npart,massarr,time,redshift,npartall);
     if(!stars) npart[4]=0;
     nskip=npart[1]+npart[2]+npart[3];
     cout << "Reading " << npart[0] << " particles + " << npart[4] << " stars " << endl;
     cout << "Distributing at " << npstart << " and " << nstarstart << endl;
     gadget_find_block(infile,"POS ");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       infile >> p[npstart+m].x >> p[npstart+m].y >> p[npstart+m].z;
     if(stars)
       {
       cout << "Skipping " << nskip << " Positions ..." << endl;
       if(nskip>0) infile.skip(nskip*4*3);
       if(npart[4]>0)
         {
         for (int m=0; m<npart[4]; ++m)
           infile >> p[m+nstarstart].x >> p[m+nstarstart].y >> p[m+nstarstart].z;
         }
       }

     for (int m=0; m<npart[0]; ++m)
        p[m+npstart].type=0;

     gadget_find_block(infile,"HSML");
     infile.skip(4);
     for (int m=0; m<npart[0]; ++m)
       {
          infile >> p[m+npstart].r;
          p[m+npstart].r *= hsml_fac;
       }
     if (lable_col_nr == 0)
       {
          for (int m=0; m<npart[0]; ++m)
            p[m+npstart].T=1.0;
       }
     else
       {
         gadget_find_block(infile,lable_col);
         infile.skip(4);
         for (int m=0; m<npart[0]; ++m)
           {
           infile >> p[m+npstart].T;
           if(log_col) p[m+npstart].T = log(p[m+npstart].T);
           mintmp=min(mintmp,p[m+npstart].T);
           maxtmp=max(maxtmp,p[m+npstart].T);
           }
       }

     if (lable_int_nr == 0)
       {
          for (int m=0; m<npart[0]; ++m)
            p[m+npstart].I=1.0;
       }
     else
       {
          gadget_find_block(infile,lable_int);
          infile.skip(4);
          for (int m=0; m<npart[0]; ++m)
            {
            infile >> p[m+npstart].I;
            if(log_int) p[m+npstart].I = log(p[m+npstart].I);
            minint=min(minint,p[m+npstart].I);
            maxint=max(maxint,p[m+npstart].I);
            }
       }

     if(stars)
       {
       if(npart[4]>0)
         {
	 if (star_hsml<=0)
	   {
            gadget_find_block(infile,"HSMS");
            infile.skip(4);
	   }
         for (int m=0; m<npart[4]; ++m)
	   {
             p[m+nstarstart].type=1;
             p[m+nstarstart].T=1;
             if (star_hsml<=0)
	       {
                 infile >> p[m+nstarstart].r;
	       }
             else
	       {
                 p[m+nstarstart].r=star_hsml;
	       }
	   }
         if (lable_int_nr == 0)
           {
            for (int m=0; m<npart[4]; ++m)
              p[m+nstarstart].I=1.0;
           }
         else
           {
              gadget_find_block(infile,stars_int);
              infile.skip(4);
              for (int m=0; m<npart[4]; ++m)
                {
                infile >> p[m+nstarstart].I;
                if(log_int) p[m+nstarstart].I = log(p[m+nstarstart].I);
                minint_stars=min(minint_stars,p[m+nstarstart].I);
                maxint_stars=max(maxint_stars,p[m+nstarstart].I);
                }
           }
         if (stars_col_nr == 0)
           {
              for (int m=0; m<npart[0]; ++m)
                p[m+npstart].I=1.0;
           }
         else
           {
             gadget_find_block(infile,stars_col);
             infile.skip(4);
             for (int m=0; m<npart[4]; ++m)
               {
               infile >> p[m+nstarstart].T;
               if(log_stars_col) p[m+nstarstart].T = log(p[m+nstarstart].T);
               mintmp_stars=min(mintmp_stars,p[m+nstarstart].T);
               maxtmp_stars=max(maxtmp_stars,p[m+nstarstart].T);
               }
           }
	 }
       }

     npstart+=npart[0];
     nstarstart+=npart[4];
    }

  if(minval_col!=maxval_col) 
    {
    mintmp=minval_col;
    maxtmp=maxval_col;
    }

  if(minval_int!=maxval_int) 
    {
    minint=minval_int;
    maxint=maxval_int;
    }

  if(minval_stars_col!=maxval_stars_col) 
    {
    mintmp_stars=minval_stars_col;
    maxtmp_stars=maxval_stars_col;
    }

  if(minval_stars_int!=maxval_stars_int) 
    {
    minint_stars=minval_stars_int;
    maxint_stars=maxval_stars_int;
    }

  cout << "Gas particles: " << endl;
  cout << "Color Range:     " << mintmp << " (min) , " << 
                                 maxtmp << " (max) " << endl;
  cout << "Intensity Range: " << minint << " (min) , " << 
                                 maxint << " (max) " << endl;
  cout << "Star particles: " << endl;
  cout << "Color Range:     " << mintmp_stars << " (min) , " << 
                                 maxtmp_stars << " (max) " << endl;
  cout << "Intensity Range: " << minint_stars << " (min) , " << 
                                 maxint_stars << " (max) " << endl;

  for (int m=0; m<np+nstar; ++m)
    {
    if(p[m].type==0)
      {
        if(lable_col_nr != 0) p[m].T = (p[m].T-mintmp)/(maxtmp-mintmp);
        if(lable_int_nr != 0) p[m].I = (p[m].I-minint)/(maxint-minint);
      }
    else
      {
        if(stars_col_nr != 0) p[m].T = (p[m].T-mintmp_stars)/(maxtmp_stars-mintmp_stars);
        if(stars_int_nr != 0) p[m].I = (p[m].I-minint_stars)/(maxint_stars-minint_stars);
      }
    }

  double fov = params.find<double>("fov",45); //in degrees
  double fovfct = tan(fov*0.5*Pi/180.);

  COLOURMAP amap,emap;

    COLOUR c0,c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12,c13,c14,c15;
    c0=COLOUR(0,0,0);              // black
    c1=COLOUR(0,0,0.999);          // blue
    c2=COLOUR(0,0.999,0);          // green
    c3=COLOUR(0,0.999,0.999);      // cyan
    c4=COLOUR(0.999,0,0);          // red
    c5=COLOUR(0.999,0,0.999);      // magenta
    c6=COLOUR(0.999,0.999,0);      // yellow
    c7=COLOUR(0.66,0.66,0.66);     // light gray
    c8=COLOUR(0.33,0.33,0.33);     // dark gray
    c9=COLOUR(0,0,0.5);            // dark blue
    c10=COLOUR(0,0.5,0);           // dark green
    c11=COLOUR(0,0.5,0.5);         // dark cyan
    c12=COLOUR(0.5,0,0);           // dark red
    c13=COLOUR(0.5,0,0.5);         // dark magenta
    c14=COLOUR(0.5,0.5,0);         // dark yellow
    c15=COLOUR(0.999,0.999,0.999); // white

  switch(colors)
    {
    case 0:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY( 0,.5,c4,c7));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,1.,c7,c1));
      emap=amap;
      break;
    case 1:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY( 0,.2,c4,c2));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.2,.5,c2,c1));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,.65,c1,c3));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.65,.8,c3,c5));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.8,1.,c5,c15));
      emap=amap;
      break;
    case 2:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.33,COLOUR(0.00,0.00,1.00),COLOUR(1.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(1.00,0.66,0.66),COLOUR(0.50,1.00,0.50)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,1.00,0.50),COLOUR(0.00,0.66,0.33)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.33),COLOUR(1.00,0.00,0.00)));
      emap=amap;
      break;
    case 3:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.0,.5,COLOUR(0.0,0.0,1.0),COLOUR(0.5,1.0,0.5)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.5,1.,COLOUR(0.5,1.0,0.5),COLOUR(1.0,0.0,0.0)));
      emap=amap;
      break;
    case 4:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.16,COLOUR(0.00,0.00,0.00),COLOUR(0.00,0.00,0.99)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.16,.33,COLOUR(0.00,0.00,0.99),COLOUR(0.99,0.66,0.99)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,COLOUR(0.99,0.66,0.99),COLOUR(0.50,0.99,0.99)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,COLOUR(0.50,0.99,0.99),COLOUR(0.00,0.66,0.66)));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,1.0,COLOUR(0.00,0.66,0.66),COLOUR(0.99,0.00,0.00)));
      emap=amap;
      break;
    case 5:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.16,c0,c9));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.16,.33,c9,c1));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.50,c1,c4));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.66,c4,c2));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.66,.83,c2,c6));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.83,1.0,c6,c15));
      emap=amap;
      break;
    case 6:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.07,c0,c9));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.07,.13,c9,c1));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.13,.20,c1,c12));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.20,.27,c12,c13));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.27,.33,c13,c10));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.33,.40,c10,c4));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.40,.47,c4,c8));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.47,.53,c8,c11));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.53,.60,c11,c5));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.60,.67,c5,c14));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.67,.73,c14,c2));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.73,.80,c2,c7));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.80,.87,c7,c3));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.87,.93,c3,c6));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.93,1.0,c6,c15));
      emap=amap;
      break;
    case 7:
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.00,.25,c0,c5));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.25,.50,c5,c4));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.50,.75,c4,c6));
      amap.Add_Entry(new LINEAR_CMAP_ENTRY(.75,1.0,c6,c15));
      emap=amap;
      break;
    }

  bool gas = params.find<bool>("gas",true);
  float zmaxval = params.find<float>("zmax",1.e23);
  float zminval = params.find<float>("zmin",0.0);
  float grayabsorb = params.find<float>("gray_absorbtion",0.2);

#ifdef GEOMETRY_FILE
  for (int m=0; m<np+nstar; ++m)
    {
       p_orig[m].x=p[m].x;
       p_orig[m].y=p[m].y;
       p_orig[m].z=p[m].z;
       p_orig[m].r=p[m].r;
       p_orig[m].ro=p[m].ro;
       p_orig[m].I=p[m].I;
       p_orig[m].T=p[m].T;
       p_orig[m].type=p[m].type;
    }
  float cam_x,cam_y,cam_z,lat_x,lat_y,lat_z,sky_x,sky_y,sky_z;
  string geometry_file = params.find<string>("geometry_file");
  string line;
  ifstream inp(geometry_file.c_str());
  int linecount=10000;
  while (getline(inp, line))
    {
      for (int m=0; m<np+nstar; ++m)
        {
           p[m].x=p_orig[m].x;
           p[m].y=p_orig[m].y;
           p[m].z=p_orig[m].z;
           p[m].r=p_orig[m].r;
           p[m].ro=p_orig[m].ro;
           p[m].I=p_orig[m].I;
           p[m].T=p_orig[m].T;
           p[m].type=p_orig[m].type;
        }
    sscanf(line.c_str(),"%f %f %f %f %f %f %f %f %f",
                      &cam_x,&cam_y,&cam_z,
	              &lat_x,&lat_y,&lat_z,
                      &sky_x,&sky_y,&sky_z);
    VECTOR campos(cam_x,cam_y,cam_z);
    VECTOR lookat(lat_x,lat_y,lat_z);
    VECTOR sky(sky_x,sky_y,sky_z);
    cout << "Camera: (" << cam_x << "," << cam_y << "," << cam_z << ")" << endl;
    cout << "Lookat: (" << lat_x << "," << lat_y << "," << lat_z << ")" << endl;
    cout << "Sky:    (" << sky_x << "," << sky_y << "," << sky_z << ")" << endl;
#else
  VECTOR campos(params.find<double>("camera_x"),
                params.find<double>("camera_y"),
                params.find<double>("camera_z"));
  VECTOR lookat(params.find<double>("lookat_x"),
                params.find<double>("lookat_y"),
                params.find<double>("lookat_z"));
  VECTOR sky(params.find<double>("sky_x",0),
                params.find<double>("sky_y",1),
                params.find<double>("sky_z",0));
#endif
  cout << "transforming" << endl;
  sky.Normalize();
  VECTOR zaxis = (lookat-campos).Norm();
  VECTOR xaxis = Cross (sky,zaxis).Norm();
  VECTOR yaxis = Cross (zaxis,xaxis);
  TRANSFORM trans;
  trans.Make_General_Transform
    (TRANSMAT(xaxis.x,xaxis.y,xaxis.z,
              yaxis.x,yaxis.y,yaxis.z,
              zaxis.x,zaxis.y,zaxis.z,
              0,0,0));
  trans.Invert();
  TRANSFORM trans2;
  trans2.Make_Translation_Transform(-campos);
  trans2.Add_Transform(trans);
  trans=trans2;
  //  trans.Add_Transform(trans2);

  for (int m=0; m<np+nstar; ++m)
    {
    VECTOR v(p[m].x,p[m].y,p[m].z);
    v=trans.TransPoint(v);
    p[m].x=v.x; p[m].y=v.y; p[m].z=v.z;
    }
#ifdef PROJECTION_OFF
    float8 dist=sqrt((campos.x-lookat.x) * (campos.x-lookat.x)
                    +(campos.y-lookat.y) * (campos.y-lookat.y)
		    +(campos.z-lookat.z) * (campos.z-lookat.z));
    float8 xfac=1./(fovfct*dist);
    cout << "Field of fiew: " << 1./xfac*2. << endl;
#else
    float8 xfac;
#endif


  for (int m=0; m<np+nstar; ++m)
    {
#ifdef PROJECTION_OFF
    p[m].x = res*.5 * (p[m].x+fovfct*dist)*xfac;
    p[m].y = res*.5 * (p[m].y+fovfct*dist)*xfac;
#else
    xfac=1./(fovfct*p[m].z);
    p[m].x = res*.5 * (p[m].x+fovfct*p[m].z)*xfac;
    p[m].y = res*.5 * (p[m].y+fovfct*p[m].z)*xfac;
#endif
    p[m].ro = p[m].r;
    p[m].r = p[m].r *res*.5*xfac;
#ifdef MINHSMLPIXEL
    if ((p[m].r <= 0.5) && (p[m].r >= 0.0))
      {
         p[m].r = 0.5;
         p[m].ro = p[m].r/res/0.5/xfac;
      }
#endif
    }

#ifdef SORTBYVAL
#ifdef REVERSESORT
  cout << "reverse sorting by value" << endl;
#else
  cout << "sorting by value" << endl;
#endif
#else
#ifdef SORTBYHSML
  cout << "sorting by hsml" << endl;
#else
  cout << "sorting by z" << endl;
#endif
#endif
  sort(&p[0], &p[np+nstar], zcmp());

#ifdef TIPSY
  cout << "plotting (adaptive smoothed tipsy like)" << endl;
#else
  cout << "plotting (raytracing)" << endl;
#endif

  float8 rfac=1.5;
  float8 powtmp = pow(float8(Pi/2),float8(1./3.));
  float8 sigma0=powtmp/sqrt(2*Pi);
  float8 i00=brightness/(2*sqrt(Pi)*powtmp);

  arr2<COLOUR> pic(res,res);
#ifdef STOP_AFTER_N
  int z_count[res][res];
#endif


  for (int m=0; m<np+nstar; ++m)
    {
    if ((m%100000)==0)
      cout << m << " of " << np+nstar << " Particles ..." << endl;

    if (p[m].z<=0) continue;
    if (p[m].z<=zminval) continue;
    if (p[m].z>=zmaxval) continue;
    //        cout << zminval << " " << p[m].z << " " << zmaxval << endl;
    if (!gas and p[m].type==0) continue;
    if (!stars and p[m].type==1) continue;
#ifdef STOP_AFTER_N
    int pxt=int(p[m].x),pyt=int(p[m].y);
    if (pxt < 0 || pxt >= res) continue;  
    if (pyt < 0 || pyt >= res) continue;  
    z_count[pxt][pyt]++;
    if (z_count[pxt][pyt] > STOP_AFTER_N) continue;
#endif

    float8 temp=p[m].T;
    temp=max(float8(0.),min(float8(0.9999999),temp));
    float8 intensity=p[m].I;
    intensity=max(float8(0.),min(float8(0.9999999),intensity));
    COLOUR e=amap.Get_Colour(temp)*intensity;
    float8 r=p[m].r;
    float8 radsq = rfac*rfac*r*r;
#ifndef TIPSY
    COLOUR a;
#ifdef FULLABSROB
    a=COLOUR(1,1,1);
#else
    a=emap.Get_Colour(temp)*intensity;
#endif

    COLOUR q (e.r/(a.r+grayabsorb),e.g/(a.g+grayabsorb),e.b/(a.b+grayabsorb));

    float8 ro=p[m].ro;
    float8 sigma=ro*sigma0;
    float8 fac0=-0.5/(sigma*sigma);
    float8 i0=i00/ro;
    float8 prefac1 = ro/r*ro/r*fac0;
    float8 prefac2 = -0.5*i0;
#endif

    float8 posx=p[m].x, posy=p[m].y;
    int minx = max(int(posx-rfac*r),0), maxx = min(int(posx+rfac*r+1),res),
        miny = max(int(posy-rfac*r),ycut0), maxy = min(int(posy+rfac*r+1),ycut1);
    for (int y=miny; y<maxy; ++y)
      {
      float8 ysq=(y-posy)*(y-posy);
      for (int x=minx; x<maxx; ++x)
        {
        float8 dsq = (x-posx)*(x-posx) + ysq;
        if (dsq<radsq)
          {
#ifdef TIPSY
	    float8 uu=max(float8(0.),min(float8(1.0),sqrt(dsq/radsq)));
            float8 lfac=0.0;
            if(uu < 0.5) lfac=1.-6.*uu*uu*(1.-uu); 
            else         lfac=2*(1-uu)*(1-uu)*(1-uu);
	    pic[x][y].r = min(e.r * lfac + (1 - lfac) * pic[x][y].r,float8(1.0));
	    pic[x][y].g = min(e.g * lfac + (1 - lfac) * pic[x][y].g,float8(1.0));
	    pic[x][y].b = min(e.b * lfac + (1 - lfac) * pic[x][y].b,float8(1.0));
	    //	    	    pic[x][y].r = e.r;
	    //	    	    pic[x][y].g = e.g;
	    //	    	    pic[x][y].b = e.b;
#else
          float8 fac = prefac2*xexp(prefac1*dsq);
          pic[x][y].r = q.r+(pic[x][y].r-q.r)*xexp(fac*a.r);
          pic[x][y].g = q.g+(pic[x][y].g-q.g)*xexp(fac*a.g);
          pic[x][y].b = q.b+(pic[x][y].b-q.b)*xexp(fac*a.b);
#endif
          }
        }
      }
    }

  bool colbar = params.find<bool>("colourbar",false);
  if (colbar)
    {
      cout << "adding colour bar ..." << endl;
      for (int x=0; x<res; x++)
        {
           float8 temp=x/float8(res);
           COLOUR e=amap.Get_Colour(temp);
           for (int y=0; y<20; y++)
	     {
	        pic[x][res-1-y].r = e.r;
    	        pic[x][res-1-y].g = e.g;
	        pic[x][res-1-y].b = e.b;
	     }
        }
    }

  cout << "writing" << endl;
  int yres=ycut1-ycut0;
  const byte header[18] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    res%256, res/256, yres%256, yres/256, 24, 32 };

#ifdef GEOMETRY_FILE
  char frame_name[300];
  sprintf(frame_name,"%s_%5i.tga",outfile.c_str(),linecount);
  bofstream file(frame_name,file_is_natural);
  linecount++;
#else
  bofstream file(outfile.c_str(),file_is_natural);
#endif
  for (int m=0; m<18; ++m) file << header[m];
  for (int y=ycut0; y<ycut1; ++y)
    {
    for (int x=0; x<res; ++x)
      {
      pic[x][y].b=min(float8(1.), max(float8(0.), float8(pic[x][y].b)));
      byte pix = min(byte(255),byte(256*pic[x][y].b));
      file << pix;
      pic[x][y].g=min(float8(1.), max(float8(0.), float8(pic[x][y].g)));
      pix = min(byte(255),byte(256*pic[x][y].g));
      file << pix;
      pic[x][y].r=min(float8(1.), max(float8(0.), float8(pic[x][y].r)));
      pix = min(byte(255),byte(256*pic[x][y].r));
      file << pix;
      }
    }

#ifdef GEOMETRY_FILE
    }
#endif
  }
